-- Author: Nicholas Riley
local addonName, addon = ...
local i, k, v
local session={}
local stagger={
  loaded = false,
  debug = false,
  inCombat = false
}
local dbg
if stagger.debug then
  function dbg(...)
    print(...)
  end
else
  function dbg(...)
  end
end

local events = {}
local defaultConfig = {
	opacity = 100,
	levels = {
		Light=0,
		Medium=6,
		Heavy=11
	}
}
local backdrop = {
  -- path to the background texture
  bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background-Dark",  
  -- path to the border texture
  edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
  -- true to repeat the background texture to fill the frame, false to scale it
  tile = true,
  -- size (width or height) of the square repeating background tiles (in pixels)
  tileSize = 32,
  -- thickness of edge segments and square size of edge corners (in pixels)
  edgeSize = 32,
  -- distance from the edges of the frame to those of the background texture (in pixels)
  insets = {
    left = 11,
    right = 12,
    top = 12,
    bottom = 11
  }
}
local function deepcopy(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[deepcopy(orig_key)] = deepcopy(orig_value)
        end
        setmetatable(copy, deepcopy(getmetatable(orig)))
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end

local staggerFrame = CreateFrame('Frame', 'StaggerFrame', UIParent)
staggerFrame:Hide()
local statusbar = CreateFrame('statusBar', 'StaggerStatusBar', staggerFrame)
local status = CreateFrame('Frame', 'StaggerStatusBar_Text', staggerFrame)
local staggerAmount = CreateFrame('Frame', 'StaggerStatusBar_Amount', staggerFrame)
local parry = CreateFrame('Frame', nil, staggerFrame)
parry.icon = parry:CreateTexture('parryIcon', 'ARTWORK')
local configPanel = CreateFrame('Frame', 'StaggerConfigPanel')
configPanel.name = addonName
configPanel:Hide()

parry.icon:SetHeight(30)
parry.icon:SetWidth(30)
parry.icon:SetTexture('Interface\\Icons\\ability_monk_shuffle')

parry:SetHeight(40)
parry:SetWidth(40)
--parry:SetBackdrop(backdrop)
parry:SetPoint('TOPRIGHT', staggerFrame, 'TOPLEFT', 12, 0)
parry.icon:SetPoint('CENTER', parry, 'CENTER', 0, 0)
parry:Hide()


local function enable()
  dbg('Stagger AddOn Enabled!')
  for k, v in pairs(events) do
    staggerFrame:RegisterEvent(k);
  end
end

local function disable()
  dbg('Stagger AddOn Disabled!')
  staggerFrame:Hide()
  for k, v in pairs(events) do
    staggerFrame:UnregisterEvent(k);
  end
end


local function checkbox(label, desc, handled)
  local check = CreateFrame('CheckButton', addonName .. label, configPanel, 'InterfaceOptionsCheckButtonTemplate')
  check.label = _G[check:GetName() .. 'Text']
  check.label:SetText(label)
  check.tooltipText = label
  check.tooltipRequirement = desc
  return check
end

local configTitle = configPanel:CreateFontString(addonName .. 'Title', 'ARTWORK', 'GameFontNormalLarge')
configTitle:SetPoint('TOPLEFT', 16, -16)
configTitle:SetText(addonName)



--local configTest = checkbox(addonName .. 'Test', 'test', function (self, value)
local function createSlider(text, props, name, tip, x, y, anchor)

  local slider = CreateFrame('Slider', addonName .. name, configPanel, 'OptionsSliderTemplate')
  slider:SetValueStep(1)
  slider:SetMinMaxValues(0,100)
  slider:SetOrientation('HORIZONTAL')
  slider:SetWidth(200)
  slider:SetHeight(20)
  slider:SetPoint('TOPLEFT', anchor, 'BOTTOMLEFT', x, y)
  slider.tooltipText = 'This is a tooltip'
  getglobal(slider:GetName() .. 'Low'):SetText('0')
  getglobal(slider:GetName() .. 'High'):SetText('100')

  slider:SetScript('OnShow', function (self)
    
    local i, prop, ref, configValue, configProp
    for i = 1, #props do
      if nil == configValue then
        configValue = session[props[i]]
      else
        ref = configValue
        configProp = props[i]
        configValue = ref[configProp]
      end
    end
    

    local title = self:CreateFontString(nil, 'ARTWORK', 'GameFontNormal')
    title:SetText(text .. ':')
    title:SetPoint('TOPLEFT', slider, 'TOPLEFT', 0, 16)
    title:Show()
    titleWidth = title:GetWidth()

    local value = self:CreateFontString(nil, 'ARTWORK', 'GameFontWhite')
    value:SetPoint('TOPLEFT', slider, 'TOPLEFT', titleWidth + 6, 16)
    value:Show()

    --update saved vars

    self.title = title
    self.value = value
    self.name = name
    self.configRef = ref
    self.configProp = configProp

    self:SetScript('OnValueChanged', self.updateValue)
    self:SetScript('OnShow', nil)
    dbg('upadting value', configValue)
    self:SetValue(configValue)
    self.updateValue(self, configValue)
  end)

  return slider
end

local configOpacity = createSlider('Opacity', {'opacity'}, 'OpacitySlider', 'adjust display opacity', 0, -32, configTitle) 
configOpacity.updateValue = function (self, value)
  dbg('updating opacity: ', value)
  value = math.floor(value)
  configOpacity.value:SetText(value .. '%')
  session.opacity = value
  staggerFrame.updateOpacity()
end


local configStagger = {
  Light = createSlider('Light Stagger', {'levels', 'Light'}, 'LightStagger', '% of life', 0, -32, configOpacity),
  Medium = createSlider('Medium Stagger', {'levels', 'Medium'}, 'MediumStagger', '% of life', 0, -84, configOpacity),
  Heavy = createSlider('Heavy Stagger', {'levels', 'Heavy'}, 'HeavyStagger', '% of life', 0, -132, configOpacity)
}
configStagger.update = function (self, value)
  value = math.floor(value)
  dbg('save:', self.configProp, value)
  local stype = self.configProp
  local levels = session.levels

  levels[stype] = value
  --self.configRef[stype] = value
  configStagger[stype].value:SetText(value .. '%')
  if stype == 'Light' then
    if levels.Medium <= levels.Light then
      configStagger.Medium:SetValue(levels.Light + 1)
    end
    if levels.Heavy <= levels.Medium then
      configStagger.Heavy:SetValue(levels.Medium + 1)
    end
  elseif stype == 'Medium' then
    if levels.Medium <= levels.Light then
      configStagger.Light:SetValue(levels.Medium - 1)
    end
    if levels.Medium >= levels.Heavy then
      configStagger.Heavy:SetValue(levels.Medium + 1)
    end
  else
    if levels.Heavy <= levels.Medium then
      configStagger.Medium:SetValue(levels.Heavy - 1)
    end
    if levels.Heavy <= levels.Light then
      configStagger.Light:SetValue(levels.Heavy - 2)
    end
  end
end

configStagger.Light.updateValue = configStagger.update
configStagger.Light:SetMinMaxValues(0,98)
getglobal(configStagger.Light:GetName() .. 'High'):SetText('98')

configStagger.Medium.updateValue = configStagger.update
configStagger.Medium:SetMinMaxValues(1,99)
getglobal(configStagger.Medium:GetName() .. 'Low'):SetText('1')
getglobal(configStagger.Medium:GetName() .. 'High'):SetText('99')

configStagger.Heavy.updateValue = configStagger.update
configStagger.Heavy:SetMinMaxValues(2,100)
getglobal(configStagger.Heavy:GetName() .. 'Low'):SetText('2')

staggerFrame:ClearAllPoints()
staggerFrame:SetHeight(70)
staggerFrame:SetWidth(140)
staggerFrame:SetBackdrop(backdrop)
staggerFrame:SetBackdropColor(1, 1, 1, 0.5)
staggerFrame:SetBackdropBorderColor(1, 1, 1, 0.7)
staggerFrame:SetPoint('BOTTOM', 0, 200)
staggerFrame.updateOpacity = function ()
  if not configPanel:IsShown() then return end
  dbg('Updating StaggerFrame opacity: ', session.opacity)
  staggerFrame:SetAlpha(session.opacity / 100)
end

statusbar:SetStatusBarTexture('Interface\\TargetingFrame\\UI-statusBar')
statusbar:GetStatusBarTexture():SetHorizTile(false)
statusbar:SetMinMaxValues(0, 3)
statusbar:SetValue(0)
statusbar:SetWidth(99)
statusbar:SetHeight(10)
statusbar:SetPoint('CENTER', staggerFrame, 'CENTER', 0, 0)

status:SetHeight(30)
status:SetWidth(200)
status.text = status:CreateFontString(nil, 'ARTWORK', 'GameFontHighlightSmall')
status.text:SetAllPoints()
status:SetPoint('BOTTOM', statusbar, 'CENTER', 0, -30)

staggerAmount:ClearAllPoints()
staggerAmount:SetHeight(30)
staggerAmount:SetWidth(200)
staggerAmount.text = staggerAmount:CreateFontString(nil, 'ARTWORK', 'GameFontHighlight')
staggerAmount.text:SetAllPoints()
staggerAmount.text:SetFontObject('GameFontHighlight')
staggerAmount.text:SetTextHeight(13)
staggerAmount:SetPoint('TOP', statusbar, 'CENTER', 0, 30)


local function setStatus(message)
  status.text:SetText(message)
end

local function setDamage(amount)
  local hp = UnitHealthMax('player')
  dbg('percent: %', percent, session.levels.Light)
	percent = math.ceil(amount / hp * 100 + 0.5)
	if percent >= session.levels.Heavy then
		stagger.warn.heavy()
	elseif percent >= session.levels.Medium then
		stagger.warn.medium()
	elseif percent >= session.levels.Light then
		stagger.warn.light()
  else
    staggerFrame:Hide()
    return
	end
  staggerAmount.text:SetText(percent .. '% : ' .. amount)
  stagger.update()
end

local function setFont(font)
  status.text:SetFontObject(font)
end

--PLAYER_LEAVING_WORLD
--staggerFrame:RegisterEvent('PLAYER_LOGOUT');

-- TODO: Capture Stagger type / Aura Name for dmg lookup on first tick
function events:UNIT_AURA(...)
  --local name, icon, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15 = UnitBuff('player', 'Shuffle')
  --print(p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15)
  if staggerFrame:IsShown() and 
		not UnitDebuff('player', 'Light Stagger') and
		not UnitDebuff('player', 'Medium Stagger') and
		not UnitDebuff('player', 'Heavy Stagger') then
    stagger.unset()
    staggerFrame:Hide()
  else
    stagger.update()
  end
end

function events:PLAYER_REGEN_DISABLED()
  stagger.inCombat = true
end

function events:PLAYER_REGEN_ENABLED()
  stagger.inCombat = false
end

function events:COMBAT_LOG_EVENT(...)
  local combatEvent, p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, debuff, rank, amount, p11, p12, p13, p14, p15 = ...;
  --print(combatEvent, debuff, rank, amount)
  --print(p1, p2, p3, p4, p5, p6, p7, p8, p9, p10, p11, p12, p13, p14, p15)
  if debuff ~= 'Stagger' then return end
	if not staggerFrame:IsShown() then
		staggerFrame:Show()
	end
  setDamage(amount * 10)
end


stagger.warn = {
  light = function(self)
    setStatus('Light Stagger!')
    setFont('GameFontGreen')
    statusbar:SetStatusBarColor(0,1,0)
    statusbar:SetValue(1)
  end,
  medium = function(self)
    setStatus('Medium Stagger!!')
    setFont('GameFontNormal')
    statusbar:SetStatusBarColor(1,0.8,0)
    statusbar:SetValue(2)
  end,
  heavy = function(self)
    setStatus('Heavy Stagger!!!')
    setFont('GameFontRed')
    statusbar:SetStatusBarColor(1,0,0)
    statusbar:SetValue(3)
  end
}

stagger.update = function(self)
  if staggerFrame:IsShown() 
    and UnitBuff('player', 'Shuffle') then 
    if not parry:IsShown() then parry:Show() return end
  elseif parry:IsShown() then
    parry:Hide()
  end
end

stagger.unset = function(self)
  if parry:IsShown() then
    parry:Hide()
  end
  staggerFrame:Hide()
  setStatus(0)
  setDamage(0)
  setFont('GameFontHighlightSmall')
  statusbar:SetValue(0)
end


--set global variables
function setVariables()
  dbg('Setting variables!')
  stagger.loaded = true
  stagger.inCombat = UnitAffectingCombat('player')
  if nil == staggerConfig then
    staggerConfig = deepcopy(defaultConfig)
  end
  --copy the data into a session!
  session = deepcopy(staggerConfig)
  configOpacity:SetValue(session.opacity)
end

--reset all settings to defaults
configPanel.default = function ()
  if session.opacity ~= defaultConfig.opacity then
    configOpacity:SetValue(defaultConfig.opacity)
  end
  local levels = session.levels
  for k, v in pairs(levels) do
    if v ~= defaultConfig.levels[k] then
      levels[k] = defaultConfig.levels[k]
      configStagger[k]:SetValue(levels[k])
    end
  end
end

--save all settings to global vars
configPanel.okay = function ()
  staggerConfig.opacity = configOpacity:GetValue()
  staggerConfig.levels.Light = configStagger.Light:GetValue()
  staggerConfig.levels.Medium = configStagger.Medium:GetValue()
  staggerConfig.levels.Heavy = configStagger.Heavy:GetValue()
end

--cancel settings changes - create new session from last saved
configPanel.cancel = function ()
  dbg('cancelling')
  local session = deepcopy(staggerConfig)
  local levels = session.levels
  configOpacity:SetValue(session.opacity)
  for k, v in pairs(levels) do
    configStagger[k]:SetValue(levels[k])
  end
end


--listen for events
staggerFrame:SetScript('OnEvent', function(self, event, unit, ...)
  -- if not loaded return, else
  --print(event, unit)
  if event == 'ADDON_LOADED' and unit == addonName then
    dbg('Monk Addon Loaded!')
    setVariables()
  end
  if event == 'PLAYER_ENTERING_WORLD' or
    event == 'ACTIVE_TALENT_GROUP_CHANGED' then
    local class = UnitClass('player')
    local specID = GetSpecialization()
    local gid, classSpec = GetSpecializationInfo(GetSpecialization())
    -- only load config for Monk class
    if class == 'Monk' then
      local categories = INTERFACEOPTIONS_ADDONCATEGORIES
      local interfaceExists = false
      for i = 1, #categories do
        if (categories[i].name == addonName) then
          interfaceExists = true
          break
        end
      end
      if (not interfaceExists) then 
        InterfaceOptions_AddCategory(configPanel)

        --InterfaceOptionsFrame_Show()
        --InterfaceOptionsFrame_OpenToCategory(configPanel.name)
      end
    end
    -- only run addon if Monk:Brewmaster
    if classSpec ~= 'Brewmaster' then
      if stagger.loaded then disable() end
    else
      if stagger.loaded then enable() end
    end
  end
  -- event must be combat type or on player
  if events[event] then 
    events[event](self, ...) -- call one of the functions above
  end
end)


staggerFrame:RegisterEvent('PLAYER_ENTERING_WORLD');
staggerFrame:RegisterEvent('ACTIVE_TALENT_GROUP_CHANGED');
staggerFrame:RegisterEvent('ADDON_LOADED');
staggerFrame:RegisterEvent('PLAYER_REGEN_ENABLED');
staggerFrame:RegisterEvent('PLAYER_REGEN_DISABLED');
